@@ -1,11 +1,12 @@
 #! /usr/local/bin/scotty -inf
 ##
-## Simple IP Discovering Tool.
+## Simple IP-Discovering Tool.
 ##
-## Copyright (c) 1993, 1994
-##                    J. Schoenwaelder
-##                    TU Braunschweig, Germany
-##                    Institute for Operating Systems and Computer Networks
+## Copyright (c) 1993, 1994, 1995
+##
+## J. Schoenwaelder
+## TU Braunschweig, Germany
+## Institute for Operating Systems and Computer Networks
 ##
 ## Permission to use, copy, modify, and distribute this
 ## software and its documentation for any purpose and without
@@ -16,6 +17,10 @@
 ## express or implied warranty.
 ##
 
+LoadDefaults icmp
+
+IpInit IP-Discover
+
 ##
 ## These are the global parameters that control the discovering
 ## process. They are manipulated by the "Discover Parameter" proc.
@@ -24,45 +29,9 @@
 set gridx 65
 set gridy 50
 set columns 16
-set icmp_retries 3
-set icmp_timeout 3
-set icmp_delay 5
-set max_route_len 8
 set report true
+set debug false
 
-##
-## The the defaults for the icmp command (see scotty(1)).
-##
-
-icmp -retries $icmp_retries
-icmp -timeout $icmp_timeout
-icmp -delay $icmp_delay
-
-##
-## Write a report to a log object.
-##
-
-proc writeln {{txt ""}} {
-    global report
-    static log
-
-    if {$report != "true"} {
-	catch {unset log}
-	return
-    }
-
-    if {![info exists log]} {
-	set log [ined -noupdate create LOG]
-	ined -noupdate name $log "IP Discover Log"
-    }
-
-    if {[ined -noupdate retrieve $log] == ""} {
-	set log [ined -noupdate create LOG]
-	ined -noupdate name $log "IP Discover Log"
-    }
-
-    ined append $log $txt
-}
 
 ##
 ## During our icmp fire, we build up the following tables.
@@ -79,22 +48,15 @@ proc writeln {{txt ""}} {
 ##    gateways[id]     <exists if id is (part of) a gateway>
 ##
 
-## Get the IP Address of a node. Query the name server, if the
-## address attribute is not set to something that looks like a
-## valid IP address.
+##
+## Reset the global tables. This is now called before we start anything
+## else to recover from abnormal terminated runs.
+##
 
-proc Get_IP_Address {node} {
-    if {[ined_type $node]=="NODE"} {
-	set host [ined_name $node]
-	set ip [ined_address $node]
-	if {[regexp "^\[0-9\]+\.\[0-9\]+\.\[0-9\]+\.\[0-9\]+$" $ip]>0} {
-	    return $ip
-	}
-	if {[catch {nslook $host} ip]==0} {
-	    return [lindex $ip 0]
-	}
+proc reset {} {
+    catch {
+	unset nodes networks links ids trace mask fip name address gateways
     }
-    return ""
 }
 
 ##
@@ -103,8 +65,7 @@ proc Get_IP_Address {node} {
 ##
 
 proc date {} {
-    if {[catch {exec date} date]} {set date ""}
-    return $date
+    return [getdate]
 }
 
 ##
@@ -116,7 +77,11 @@ proc date {} {
 proc create_node {ip} {
     global ids nodes address name
     set id [newid]
-    if {[catch {dns ptr $ip} ns_name]} { set ns_name $ip }
+    if {[catch {dns ptr $ip} ns_name]} {
+	if {[catch {nslook $ip} ns_name]} {
+	    set ns_name $ip
+	}
+    }
     set ns_name [lindex $ns_name 0]
     set ids($ip) $id
     set nodes($id) ""
@@ -197,7 +162,7 @@ proc netping { network } {
 	for {set a4 1} {$a4<255} {incr a4} {
 	    append hosts " $network.$a4"
 	}
-	set result [icmp $hosts]
+	set result [icmp echo $hosts]
     }
     set res ""
     foreach pr $result {
@@ -257,17 +222,17 @@ proc discover_nodes {network} {
 ##
 
 proc trace_route {iplist} {
-    global max_route_len
+    global icmp_routelength
 
     foreach ip $iplist { set routetable($ip) "" }
     set todo $iplist
     set ttl 1
 
     while {$todo != ""} {
-	if {$ttl > $max_route_len} break
+	if {$ttl > $icmp_routelength} break
 	set i 0
 	set new_todo ""
-	foreach aa [icmp -trace $ttl $todo] {
+	foreach aa [icmp trace $ttl $todo] {
 	    set ip [lindex $todo $i]
 	    set trace_ip [lindex $aa 0]
 	    set trace_time [lindex $aa 1]
@@ -354,7 +319,7 @@ proc discover_masks {} {
     foreach id [array names nodes] {
 	lappend addrs $address($id)
 	if {[llength $addrs] > 255} {
-	    foreach ipmask [eval icmp -mask $addrs] {
+	    foreach ipmask [icmp mask $addrs] {
 		set id $ids([lindex $ipmask 0])
 		set mask($id) [lindex $ipmask 1]
 		incr count
@@ -362,7 +327,7 @@ proc discover_masks {} {
 	    set addrs ""
 	}
     }
-    foreach ipmask [eval icmp -mask $addrs] {
+    foreach ipmask [icmp mask $addrs] {
 	set id $ids([lindex $ipmask 0])
 	set mask($id) [lindex $ipmask 1]
 	incr count
@@ -373,6 +338,44 @@ proc discover_masks {} {
 
 
 
+##
+## ===========================================================================
+## ========== P H A S E   3.5 (Guess snmp parameters) ========================
+## ===========================================================================
+##
+
+proc discover_snmp_callback {id s e} {
+    global snmp
+    if {$e == "noError"} {
+	set snmp($id) [$s configure]
+    }
+    $s destroy
+}
+
+proc discover_snmp {} {
+    global nodes address snmp
+    global icmp_retries icmp_timeout
+    set start [getclock]
+    mib load rfc1213.mib
+    foreach id [array names nodes] {
+        set ip $address($id)
+	if {[catch {snmp session -address $ip \
+		-retries $icmp_retries -timeout $icmp_timeout} s]} continue
+	if {[catch {
+	    $s get sysObjectID.0 [list discover_snmp_callback $id "%S" "%E"]
+	} msg]} {
+	    writeln "Oops: $ip get sysObjectID.0: $msg"
+	}
+        update
+    }
+    snmp wait
+    set count [llength [array names snmp]]
+    writeln "$count snmp agents queried in [expr {[getclock]-$start}] seconds."
+    flush stdout
+}
+
+
+
 ##
 ## ===========================================================================
 ## ========== P H A S E   4   (Get the ip address as returned from the nodes =
@@ -398,7 +401,7 @@ proc discover_fips {} {
 	lappend idlist $id
 	if {[llength $addrs] > 255} {
 	    set idx 0
-	    foreach ipfip [eval icmp -ttl 32 $addrs] {
+	    foreach ipfip [icmp ttl 32 $addrs] {
 		set id [lindex $idlist $idx]
 		set fip($id) [lindex $ipfip 0]
 		incr count
@@ -409,7 +412,7 @@ proc discover_fips {} {
 	}
     }
     set idx 0
-    foreach ipfip [eval icmp -ttl 32 $addrs] {
+    foreach ipfip [icmp ttl 32 $addrs] {
 	set id [lindex $idlist $idx]
 	set fip($id) [lindex $ipfip 0]
 	incr count
@@ -433,10 +436,10 @@ proc discover_fips {} {
 
 proc ip_class {network} {
     set byte_one [lindex [split $network .] 0]
-    if {[expr {$byte_one <  127}]} { return A }
-    if {[expr {$byte_one == 127}]} { return loopback }
-    if {[expr {$byte_one <  192}]} { return B }
-    if {[expr {$byte_one <  224}]} { return C }
+    if {$byte_one <  127} { return A }
+    if {$byte_one == 127} { return loopback }
+    if {$byte_one <  192} { return B }
+    if {$byte_one <  224} { return C }
     return D
 }
 
@@ -513,7 +516,7 @@ proc netmask_compare {maska maskb} {
 ##
 
 proc discover_networks {} {
-    global nodes networks ids trace mask address
+    global nodes networks ids trace mask address report
     set count 0
     set start [getclock]
 
@@ -544,7 +547,11 @@ proc discover_networks {} {
 	    if {![catch {nslook [exec hostname]} localhost]} {
 		if {[string trim $localhost] == $ip} continue
 	    }
-	    ined acknowledge "Please check the netmask $netmask for $ip on class $class network $net."
+	    set txt "Please check the netmask $netmask for $ip on class $class network $net."
+	    ined acknowledge $txt
+	    if {$report == "true"} {
+		writeln $txt
+	    }
 	    continue
 	}
 
@@ -576,7 +583,7 @@ proc discover_networks {} {
 ##
 
 proc validate_networks { net_table } {
-    global nodes address mask ids
+    global nodes address mask ids report
     upvar $net_table table
 
     # Make a majority decision if the subnet is a vaild one or
@@ -622,10 +629,19 @@ proc validate_networks { net_table } {
 	    }
 	}
 	if {$cons > $pros} {
-	    set res [ined confirm "Discard network $net?" "" \
-		     "The majority ($cons : $pros) of nodes on this" \
-		     "network have wide netmask." ]
-	    if {$res == "yes"} { unset table($net) }
+	    set res [ined confirm \
+  "The majority ($cons : $pros) of nodes on network $net have wide netmasks." \
+  "$unk nodes did not vote. Discard this network?" [list no yes]]
+	    if {$res == "yes"} { 
+		unset table($net) 
+		if {$report == "true"} {
+		    writeln "Network $net discarded ($pros, $cons, $unk)." 
+		}
+	    } else {
+		if {$report == "true"} {
+		    writeln "Network $net not discarded ($pros, $cons, $unk)."
+		}
+	    }
 	}
     }
 
@@ -653,7 +669,7 @@ proc validate_networks { net_table } {
 
 proc discover_gateways {} {
     global ids trace mask nodes address name gateways fip
-    global max_route_len
+    global icmp_routelength
     set count 0
     set start [getclock]
 
@@ -852,7 +868,7 @@ proc discover_links {} {
 ##
 
 ##
-## Forevery discovered gateway with more than one known interface,
+## For each discovered gateway with more than one known interface,
 ## merge the nodes for the interfaces to a single one. 
 ##
 
@@ -863,6 +879,12 @@ proc merge_gateways {} {
 
     if {![info exists gateways]} return
 
+    debug "** entering merge_gateways"
+    foreach id [array names gateways] {
+	if {[llength $gateways($id)] < 2} continue
+	debug "** gateways ($id) :\t $gateways($id)"
+    }
+
     # We set up a table indexed by the connected ids.
     # This is used to avoid duplicates.
 
@@ -875,61 +897,97 @@ proc merge_gateways {} {
     # to survive. Nodes with a valid domain name are prefered.
 
     foreach id [array names gateways] {
+
 	if {[llength $gateways($id)] < 2} continue
+
 	set myip [lindex $gateways($id) 0]
+	if {[info exists done($myip)]} {
+	    debug "** $myip skipped - $myip is already done"
+	    continue
+	}
 
-	if {![info exists done($myip)]} {
+	# Check if one of the foreign addresses of this gateway
+	# contains more ip address. Skip this entry if we find
+	# such an entry that also contains myip.
 
-	    if {$name($id) == $myip} {
-		set ok 0
-		foreach ip [lrange $gateways($id) 1 end] {
-		    if {$ip == "0.0.0.0"} continue
-		    set aa $ids($ip)
-		    if {$ip != $name($aa)} {
-			set ok 1
-			break
-		    }
-		}
-		if {$ok} continue
+	set skip 0
+	foreach ip [lrange $gateways($id) 1 end] {
+debug "--   checking $ip for length"
+	    if {($ip == "0.0.0.0") || (![info exists ids($ip)])} continue
+	    set aid $ids($ip)
+	    if {![info exists gateways($aid)]} continue
+	    if {[llength $gateways($aid)] > [llength $gateways($id)]} {
+		set skip 1
+		# mark this ip number so that it will not skipped
+		# because of any other reasons.
+		set force($ip) $ip  
+		break
 	    }
+	}
 
-	    set done($myip) 1
-	    incr count
+	if {$skip} {
+	    debug "** $myip skipped - using $ip instead (more interfaces)"
+	    continue
+	}
+
+	# Search throught the ip addresses of the gateway to find an
+	# address with a real name if the name of the current gateway 
+	# is just the ip address.
+
+	if {(![info exists force($myip)]) && ($name($id) == $myip)} {
+	    set skip 0
 	    foreach ip [lrange $gateways($id) 1 end] {
-
-		if {$ip == "0.0.0.0"} continue
-
-		# Merge node given by ip with the given by id. Ignore IP 
-		# addresses for which there is no id. We get such beasts
-		# when checking the returned udp address.
-
-		if {![info exists ids($ip)]} continue
-
-		set rid $ids($ip)
-		catch {unset nodes($rid)}
-		catch {unset address($rid)}
-		catch {unset name($rid)}
-
-		# Adjust the links pointing to rid.
-
-		foreach lid [array names links] {
-		    if {[lsearch $links($lid) $rid] < 0} continue
-		    set ida [lindex $links($lid) 0]
-		    set idb [lindex $links($lid) 1]
-		    if {$ida == $rid} { set ida $id }
-		    if {$idb == $rid} { set idb $id }
-
-		    # Remove the old link and create a new one
-		    # if it does not already exist.
-
-		    unset links($lid)
-		    set aa "$ida $idb"
-		    if {![info exists already($aa)]} {
-			set links($lid) $aa
-		    }
+debug "--   checking $ip for name"
+		if {($ip == "0.0.0.0") || (![info exists ids($ip)])} continue
+		set aid $ids($ip)
+		if {$ip != $name($aid)} {
+		    set skip 1
+		    break
 		}
-		set done($ip) 1
 	    }
+	    if {$skip} {
+		debug "** $myip skipped - using $ip instead (better name)"
+		continue
+	    }
+	}
+	
+	set done($myip) 1
+	incr count
+	debug "** $myip merged with [lrange $gateways($id) 1 end]"
+	foreach ip [lrange $gateways($id) 1 end] {
+	    
+	    if {$ip == "0.0.0.0"} continue
+	    
+	    # Merge node given by ip with the given by id. Ignore IP 
+	    # addresses for which there is no id. We get such beasts
+	    # when checking the returned udp address.
+	    
+	    if {![info exists ids($ip)]} continue
+	    
+	    set rid $ids($ip)
+	    catch {unset nodes($rid)}
+	    catch {unset address($rid)}
+	    catch {unset name($rid)}
+	    
+	    # Adjust the links pointing to rid.
+	    
+	    foreach lid [array names links] {
+		if {[lsearch $links($lid) $rid] < 0} continue
+		set ida [lindex $links($lid) 0]
+		set idb [lindex $links($lid) 1]
+		if {$ida == $rid} { set ida $id }
+		if {$idb == $rid} { set idb $id }
+		
+		# Remove the old link and create a new one
+		# if it does not already exist.
+		
+		unset links($lid)
+		set aa "$ida $idb"
+		if {![info exists already($aa)]} {
+		    set links($lid) $aa
+		}
+	    }
+	    set done($ip) 1
 	}
     }
 
@@ -945,11 +1003,11 @@ proc merge_gateways {} {
 ##
 
 ##
-## Talk with ined to draw the network map.
+## Talk with tkined to let him draw the network map.
 ##
 
 proc talk_to_ined {} {
-    global nodes networks links address name
+    global nodes networks links address snmp name
     set count 0
     set start [getclock]
 
@@ -958,29 +1016,30 @@ proc talk_to_ined {} {
 
     set grouplist ""
     foreach comp [ined retrieve] {
-	lappend grouplist [ined_id $comp]
+	lappend grouplist [ined id $comp]
     }
     while {[llength $grouplist] > 0} {
 	set newlist ""
 	foreach id $grouplist {
 	    set comp [ined retrieve $id]
-	    set type [ined_type $comp]
-	    set id [ined_id $comp]
+	    set type [ined type $comp]
+	    set id [ined id $comp]
 	    switch $type {
 		NODE {
-		    set id_by_addr([Get_IP_Address $comp]) $id
+		    set addr [GetIpAddress $comp]
+		    set id_by_addr($addr) $id
 		}
 		NETWORK {
-		    set id_by_addr([ined_address $comp]) $id
+		    set id_by_addr([ined address $comp]) $id
 		}
 		LINK {
-		    set ida [ined_ida $comp]
-		    set idb [ined_idb $comp]
+		    set ida [ined src $comp]
+		    set idb [ined dst $comp]
 		    set aa "$ida $idb"
 		    set id_by_link($aa) $id
 		}
 		GROUP {
-		    lappend newlist [ined_member $comp]
+		    lappend newlist [ined member $comp]
 		}
 	    }
 	}
@@ -1001,6 +1060,9 @@ proc talk_to_ined {} {
 	    ined -noupdate name $ined_id($id) "[lindex $name($id) 0]"
 	    ined -noupdate label $ined_id($id) name
 	    ined -noupdate move $ined_id($id) [nextx] [nexty]
+	    if {[info exists snmp($id)]} {
+		ined -noupdate attribute $ined_id($id) "SNMP:Config" $snmp($id)
+	    }
 	    set id_by_addr($ip) $ined_id($id)
 	    incr count
 	}
@@ -1072,31 +1134,36 @@ proc talk_to_ined {} {
 ##
 
 proc "Discover IP Network" {list} {
+
     global nodes networks links ids trace mask fip name address gateways
     static nets
-    if ![info exists nets] { set nets "" }
+
+    reset
+
+    if {![info exists nets]} { set nets "" }
 
     set given_nets ""
     foreach comp $list {
-	if {[ined_type $comp] == "NETWORK"} {
-	    set ip [lindex [ined_address $comp] 0]
+	if {[ined type $comp] == "NETWORK"} {
+	    set ip [lindex [ined address $comp] 0]
 	    lappend given_nets $ip
 	}
     }
-
     if {$given_nets != ""} {
-	set nets $given_nets
+	set nets [list $given_nets]
     }
 
-    set res [ined request "Welcome to the Network Discovering Tool." "" \
-	     "Enter the IP Numbers of the networks of interest:" \
-	     "{{} {$nets}}"]
-    if {$res == ""} return
+    set result [ined request "Welcome to the Network Discovering Tool." "" \
+		  "Enter the IP Numbers of the networks of interest:" \
+		  [list [list "" $nets]] [list discover cancel] ]
+
+    if {[lindex $result 0] == "cancel"} return
 
     # remove trailing .0's
 
     set tmp ""
-    foreach net [lindex $res 0] {
+    foreach net [lindex $result 1] {
+	if {[ip_class $net] == "loopback"} continue
 	set dots [llength [split $net .]]
 	if {$dots == 4} {
 	    regsub {(\.0+)+$} $net "" net
@@ -1108,13 +1175,14 @@ proc "Discover IP Network" {list} {
     set hostname [exec hostname]
     set ip [nslook $hostname]
     writeln [date]
-    writeln "discover $nets from $hostname \[$ip\]"
+    writeln "Discover $nets from $hostname \[$ip\]"
     set start [getclock]
 
     foreach network $nets { discover_nodes $network }
     if {[info exists nodes]} {
 	discover_traces
 	discover_masks
+	discover_snmp
 	discover_fips
 	discover_networks
 	discover_gateways
@@ -1125,7 +1193,7 @@ proc "Discover IP Network" {list} {
 	catch {unset gateways}
     }
 
-    writeln "discover finished in [expr {[getclock]-$start}] seconds"
+    writeln "Discover finished in [expr {[getclock]-$start}] seconds"
     writeln
 }
 
@@ -1136,38 +1204,52 @@ proc "Discover IP Network" {list} {
 ##
 
 proc "Discover Route" {list} {
+
     global nodes networks links ids trace mask fip name address gateways
     static ips
 
-    if ![info exists ips] { set ips "" }
+    reset
+
+    if {![info exists ips]} { set ips "" }
 
     set given_ips ""
     foreach comp $list {
-	if {[ined_type $comp] == "NODE"} {
-            lappend given_ips [Get_IP_Address $comp]
+	if {[ined type $comp] == "NODE"} {
+            lappend given_ips [GetIpAddress $comp]
 	}
     }
-
     if {$given_ips != ""} {
-	set ips $given_ips
+	set ips [list $given_ips]
     }
 
-    if {![info exists nodes]} {
-	set ips [ined request "Welcome to the Route Discovering Tool." "" \
-		 "Enter the IP Numbers of the destination hosts:" \
-                 "{{} {$ips}}"]
-	foreach ip [lindex $ips 0] {
-	    if {![regexp "^\[0-9\]+\.\[0-9\]+\.\[0-9\]+\.\[0-9\]+$" $ip]} {
-		set host $ip
-		if {[catch {nslook $host} ip]} {
-		    ined acknowledge "Can not lookup IP address for $host."
-		    continue
-		}
-		set ip [lindex $ip 0]
+    set ips [ined request "Welcome to the Route Discovering Tool." "" \
+	       "Enter the IP Numbers of the destination hosts:" \
+               [list "{} $ips"] [list discover cancel] ]
+
+    if {[lindex $ips 0] == "cancel"} return
+    
+    # remove duplicate ip addresses, expand domain names and 
+    # create target nodes
+
+    set new_ips ""
+    foreach ip [lindex $ips 1] {
+
+	if {[lsearch $new_ips $ip] >= 0} continue
+	lappend new_ips $ip
+
+	if {![regexp "^\[0-9\]+\.\[0-9\]+\.\[0-9\]+\.\[0-9\]+$" $ip]} {
+	    set host $ip
+	    if {[catch {nslook $host} ip]} {
+		ined acknowledge "Can not lookup IP address for $host."
+		continue
 	    }
-	    create_node $ip
+	    set ip [lindex $ip 0]
 	}
+
+	create_node $ip
     }
+    set ips $new_ips
+    if {$ips == ""} return
 
     set hostname [exec hostname]
     set ip [nslook $hostname]
@@ -1178,6 +1260,7 @@ proc "Discover Route" {list} {
     if {[info exists nodes]} {
 	discover_traces
 	discover_masks
+	discover_snmp
 	discover_fips
 	discover_networks
 	discover_gateways
@@ -1197,31 +1280,38 @@ proc "Discover Route" {list} {
 ## Set some constants that will control the discovery process.
 ##
 
-proc "set parameter" {list} {
+proc "Set Parameter" {list} {
     global icmp_retries icmp_timeout icmp_delay gridx gridy columns
-    global max_route_len
+    global icmp_routelength
+    global email_trace
     global report
+    global debug
 
-    set result [ined request "IP Discover Parameter" \
-	"{ {# of ICMP retries:} $icmp_retries scale 1 10} \
-         { {ICMP timeout \[s\]:} $icmp_timeout scale 1 42} \
-         { {Delay between ICMP packets \[ms\]:} $icmp_delay scale 1 100} \
-         { {Max. length of a route:} $max_route_len scale 1 30} \
-         { {Horizontal grid space:} $gridx scale 20 160 } \
-         { {Vertical grid space:} $gridy scale 20 160 } \
-         { {Nodes per row:} $columns scale 10 40} \
-         { {Write Report:} $report radio true false} "]
+    set result [ined request "IP-Discover Parameter" \
+	[list [list "# of ICMP retries:" $icmp_retries scale 1 10] \
+          [list "ICMP timeout \[s\]:" $icmp_timeout scale 1 42] \
+          [list "Delay between ICMP packets \[ms\]:" $icmp_delay scale 1 100] \
+          [list "Max. length of a route:" $icmp_routelength scale 1 42] \
+          [list "Horizontal grid space:" $gridx scale 20 160] \
+          [list "Vertical grid space:" $gridy scale 20 160] \
+          [list "Nodes per row:" $columns scale 10 40] \
+          [list "Email Discover Routes:" $email_trace radio true false] \
+          [list "Write Report:" $report radio true false] \
+          [list "Debug Mode:" $debug radio true false] ] \
+        [list "set values" cancel] ]
 
-    if {$result==""} return
+    if {[lindex $result 0] == "cancel"} return
 
-    set icmp_retries  [lindex $result 0]
-    set icmp_timeout  [lindex $result 1]
-    set icmp_delay    [lindex $result 2]
-    set max_route_len [lindex $result 3]
-    set gridx         [lindex $result 4]
-    set gridy         [lindex $result 5]
-    set columns       [lindex $result 6]
-    set report        [lindex $result 7]
+    set icmp_retries     [lindex $result  1]
+    set icmp_timeout     [lindex $result  2]
+    set icmp_delay       [lindex $result  3]
+    set icmp_routelength [lindex $result  4]
+    set gridx            [lindex $result  5]
+    set gridy            [lindex $result  6]
+    set columns          [lindex $result  7]
+    set email_trace      [lindex $result  8]
+    set report           [lindex $result  9]
+    set debug            [lindex $result 10]
 
     icmp -retries $icmp_retries
     icmp -timeout $icmp_timeout
@@ -1232,153 +1322,385 @@ proc "set parameter" {list} {
 ## Find RPC Server by querying the portmapper.
 ##
 
-proc "Find RPC Server" {list} {
-    if {[catch {open /etc/rpc} fh]} {
-	ined acknowledge "Can not open /etc/rpc."
-        return
-    }
-    
+proc "Show RPC Server" {list} {
+
     set service_list ""
-    while {! [eof $fh]} {
-	gets $fh line
-	if {![string match "#*" $line]} {
-	    lappend service_list [lindex $line 0]
-	}
+    foreach rpc [netdb sunrpcs] {
+	lappend service_list [lindex $rpc 0]
     }
-    close $fh
 
-    set service [ined list "Select the RPC service of interest" \
-		 [lsort $service_list]]
-    if {$service==""} return
+    set service [ined list "Select the RPC service of interest:" \
+		 [lsort $service_list] [list select cancel]]
+    if {[lindex $service 0] == "cancel"} return
 
+    set service [lindex [lindex $service 1] 0]
     set hostname [exec hostname]
     set ip [nslook $hostname]
-    writeln [date]
-    writeln "find RPC server $service from $hostname \[$ip\]"
+    writeln "Show RPC server $service from $hostname \[$ip\]:"
     set start [getclock]
 
-    set found ""
     foreach comp $list {
-        if {[ined_type $comp]=="NODE"} {
-            set id [ined_id $comp]
-	    set host [ined_name $comp]
-	    set ip [Get_IP_Address $comp]
+        if {[ined type $comp]=="NODE"} {
+            set id [ined id $comp]
+	    set host [ined name $comp]
+	    set ip [GetIpAddress $comp]
 	    set color($id) [ined color $id]
 	    set label($id) [ined label $id]
-	    set rtt [lindex [icmp $ip] 0]
+	    set rtt [lindex [icmp echo $ip] 0]
 	    if {[lindex $rtt 1]<=0} continue
-	    if {![catch {eval rpc info $ip} portmapinfo]} {
+	    if {![catch {eval sunrpc info $ip} portmapinfo]} {
 		foreach serv $portmapinfo {
 		    if {[lsearch $serv $service]>=0} {
-			lappend found $id
-			ined label $id text "$service"
-			writeln [format "%-32s \[%s\]" $host $ip]
+			write   [format "%-32s \[" $host ]
+			write   $ip "IpFlash $ip"
+			writeln "\]"
 			break
 		    }
 		}
 	    }
 	}
     }
-
-    writeln "find finished in [expr {[getclock]-$start}] seconds"
     writeln
-
-    if {$found==""} return
-
-    foreach id $found {
-	ined -noupdate color $id Blue
-	ined -noupdate label $id text "$service"
-    }
-
-    sleep 6
-
-    foreach id $found {
-	ined -noupdate color $id $color($id)
-	eval ined -noupdate label $id $label($id)
-    }
-
 }
 
 ##
 ## Find TCP Server by connecting to a well known TCP port.
 ##
 
-proc "Find TCP Server" {list} {
-    if {[catch {open /etc/services} fh]} {
-	ined acknowledge "Can not open /etc/services."
-        return
-    }
-    
-    set service_list "X11"
-    while {! [eof $fh]} {
-	gets $fh line
-	if {![string match "#*" $line]} {
-	    if {[regexp "\[0-9\]*/tcp" $line]>0} {
-		lappend service_list [lindex $line 0]
-	    }
-	}
-    }
-    close $fh
+proc "Show TCP Server" {list} {
 
-    set service [ined list "Select the RPC service of interest" \
-		 [lsort $service_list]]
-    if {$service==""} return
+    set service [IpService tcp]
+
+    if {$service == ""} return
 
     set hostname [exec hostname]
     set ip [nslook $hostname]
-    writeln [date]
-    writeln "find TCP server $service from $hostname \[$ip\]"
+    set name [lindex $service 0]
+    set port [lindex $service 1]
+    writeln "Show TCP server $name ($port) from $hostname \[$ip\]:"
     set start [getclock]
 
-    set found ""
     foreach comp $list {
-        if {[ined_type $comp]=="NODE"} {
-            set id [ined_id $comp]
-	    set host [ined_name $comp]
-	    set ip [Get_IP_Address $comp]
+        if {[ined type $comp] == "NODE"} {
+            set id [ined id $comp]
+	    set host [ined name $comp]
+	    set ip [GetIpAddress $comp]
 	    set color($id) [ined color $id]
 	    set label($id) [ined label $id]
-	    set rtt [lindex [icmp $ip] 0]
-	    if {[lindex $rtt 1]<=0} continue
 
-	    if {$service == "X11" } {
-		set name 6000
-	    } else {
-		set name [lindex $service 0]
+	    if {[lindex [lindex [icmp echo $ip] 0] 1] <= 0} {
+		ined acknowledge "Skipping $host (currently unreachable)"
+		continue
 	    }
-	    if {![catch {connect $ip $name} f]} {
-		lappend found $id
-		shutdown $f all
-		ined label $id text "$service"
-		writeln [format "%-32s \[%s\]" $host $ip]
+
+	    if {![catch {tcp connect $ip $port} f]} {
+		tcp shutdown $f all
+		tcp close $f
+		write   [format "%-32s \[" $host ]
+		write   $ip "IpFlash $ip"
+		writeln "\]"
 	    }
 	}
     }
-
-    writeln "find finished in [expr {[getclock]-$start}] seconds"
     writeln
+}
 
-    if {$found==""} return
+##
+## Here are the beginnings of the email tools.
+##
 
-    foreach id $found {
-	ined -noupdate color $id Blue
-	ined -noupdate label $id text "$service"
+set last_email ""
+set max_level 12
+set email_trace false
+
+##
+## Create a new icon for this host (if we did not have it yet)
+##
+
+proc ined_create {host user} {
+    global nodes networks links ids trace mask fip name address gateways
+    global hosts email_trace
+
+    if {![info exists hosts($host)]} {
+	set hosts($host) $host
     }
 
-    sleep 6
+    if {$email_trace == "true"} {
+	set hostname [exec hostname]
+	set ip [nslook $hostname]
+	set ips [lindex [nslook $host] 0]
+	create_node $ips
+	writeln [date]
+	writeln "discover route to $ips from $hostname \[$ip\]"
+	set start [getclock]
+	
+	if {[info exists nodes]} {
+	    discover_traces
+	    discover_masks
+	    discover_snmp
+	    discover_fips
+	    discover_networks
+	    discover_gateways
+	    discover_links
+	    merge_gateways
+	    talk_to_ined
+	    unset nodes networks links ids trace mask fip name address
+	    catch {unset gateways}
+	}
+	
+	writeln "discover finished in [expr {[getclock]-$start}] seconds"
+	writeln
+    } else {
+	set id [ined -noupdate create NODE]
+	ined -noupdate name $id $host
+	ined -noupdate attribute $id email "$user@$host"
+	ined -noupdate label $id email
+	ined -noupdate move $id [nextx] [nexty]
+    }
 
-    foreach id $found {
-	ined -noupdate color $id $color($id)
-	eval ined -noupdate label $id $label($id)
+}
+
+##
+## Print n spaces to stdout.
+##
+
+proc space {n} {
+    for {set i 0} {$i < $n} {incr i} { puts -nonewline "  " }
+}
+
+##
+## Read an answer from an SMTP server and return the whole stuff
+## in from of a tcl list (with all newline characters removed).
+##
+
+proc read_answer {f} {
+    set line [string trim [gets $f] "\r"]
+    set answer [list $line]
+    while {[string index $line 3] == "-"} {
+        set line [string trim [gets $f] "\r"]
+	lappend answer $line
+	flush stdout
+    }
+    return $answer
+}
+
+##
+## Contact the smtp daemon at a given host and check if it
+## can verify the given user.
+##
+
+proc test_user_host {user host level} {
+
+    if {$level > 20} {
+        space $level; writeln "-> a loop? aborting after 20 hops"
+	return 0
+    }
+
+    ined_create $host $user
+
+    if {[catch {tcp connect $host smtp} f]} {
+	space $level; writeln "-> unable to connect to $host"
+	return 0
+    }
+    if {[catch {read_answer $f}]} {
+	return 0
+    }
+
+    if {![catch {dns ptr [nslook [exec hostname]]} hostname]} {
+	puts $f "helo $hostname"
+	if {[catch {read_answer $f}]} {
+	    return 0
+	}
+    }
+
+    if {[catch {puts $f "vrfy $user"}]} {
+	space $level; writeln "-> unable to write to $host"
+	return 0
+    }
+
+    if {[catch {read_answer $f} answer]} {
+	return 0
+    }
+    foreach addr $answer {
+
+	set code [string range $addr 0 2]
+	if {$code != 250} {
+	    space $level; writeln "$user@$host"
+	    space $level; writeln "-> $addr"
+	    set ok 0
+	    continue
+	}
+
+	# split addr into an email address and an optional name
+
+	regexp -nocase {<.*>} $addr email
+	if {![info exist email]} { set email [string range $addr 4 end] }
+	set email [string trim $email "<>"]
+	regexp -nocase {\(.*\)} $email name
+	regsub -nocase {\(.*\)} $email "" email
+	set email [string trim $email]
+
+	debug "** got email address $email"
+
+	if {![info exists name]} {
+	    set name [string range $addr 4 end]
+	    regsub -nocase {<.*>} $name "" name
+	}
+	set name [string trimleft  $name " ("]
+	set name [string trimright $name " )"]
+
+	# Check for host names that contain dots.
+	if {[string match {*@*.*} $email]} {
+	    set ok [test_email $email [expr {$level+1}] $name]
+
+	# Hostnames without dots get expanded magically.
+	} elseif {[string match {*@*} $email]} {
+	    set us [lindex [split $email @] 0]
+            set nl [split [lindex [split $email @] 1] .]
+            set ol [split $host .]
+
+	    set len [llength $ol]
+	    set ok 0
+	    for {set i 0}  {($i < $len) && ($ok == 0)} { incr i} {
+		set hn "$nl [lrange $ol $i end]"
+		set hn [join $hn "."]
+
+		# Do we have a A or a MX record for this?
+
+		if {[catch {dns address $hn} res]} {
+		    if {[catch {dns mx $hn} res]} continue
+		}
+		if {$res == ""} continue
+
+		set email "$us@$hn"
+		debug "** expanding with domain name extensions to $email"
+		set ok [test_email $email [expr {$level+1}] $name]
+	    }
+
+	# OK, no @. We believe that we have found it.
+	} else {
+	    space [expr {$level+1}]
+	    if {$name != ""} {
+		writeln "$email ($name)"
+	    } else {
+		writeln "$email"
+	    }
+	    space [expr {$level+1}]; writeln "-> ok"
+	    set ok 1
+	}
+
+	unset name
+    }
+
+    tcp close $f
+
+    return $ok
+}
+
+##
+## Test a given email. First, split it in the user and host part.
+## See if we have a local address or if we can resolve the hostname
+## to an IP address. If this fails, check if we can get an MX record
+## for the address.
+##
+
+proc test_email {email {level 0} {name ""}} {
+
+    global last_email max_level
+
+    # Loop detection.
+    if {($last_email == $email) || ($level > $max_level)} {
+	space $level
+	if {$name != ""} {
+	    writeln "$email ($name)"
+	} else {
+	    writeln "$email"
+	}
+	space $level; writeln "-> ok (possible loop detected)"
+	return 0
+    }
+    set last_email $email
+
+    # Say where we are.
+    space $level
+    if {$name != ""} {
+	writeln "$email ($name)"
+    } else {
+	writeln "$email"
+    }
+
+    set email [split  $email @]
+    set user  [lindex $email 0]
+    set host  [lindex $email 1]
+
+    ## Use localhost if we operate local.
+    if {$host == ""} { 
+	debug "** testing localhost"
+	return [test_user_host $user localhost $level]
+    }
+
+    ## expand the host if we get an MX record for it
+    if {![catch {dns mx $host} res]} {
+	set ok 0
+	foreach mx $res {
+	    if {[lindex $mx 0] == $host} break
+	    debug "** expanding based on MX record to [lindex $mx 0]"
+	    set ok [test_email "$user@[lindex $mx 0]" [expr {$level+1}]]
+	    if {$ok} break
+	}
+	return $ok
+    }
+
+    ## check if we have a real host name. convert it back into
+    ## a fully qualified hostname if possible.
+    if {![catch {dns address $host} res]} {
+	if {$res != ""} {
+	    set ok 0
+	    foreach aa $res {
+		if {![catch {dns ptr $aa} aa]} { set host $aa }
+		debug "** testing fully qualified host address $aa"
+		set ok [test_user_host $user $host $level]
+		if {$ok} break
+	    }
+	    return $ok
+        }
+    }
+
+    space $level; writeln "-> probably a bad email address"
+    return 0
+}
+
+##
+## Here is the main routine.
+##
+
+proc "Resolve Email" { list } {
+
+    static email
+
+    reset
+
+    if {![info exists email]} { set email "" }
+
+    set result [ined request "Please enter an email address to discover:" \
+		  [list [list Email: $email]] [list discover cancel] ]
+
+    if {[lindex $result 0] == "cancel"} return
+
+    set email [lindex $result 1]
+
+    foreach addr $email {
+	test_email $addr
     }
 }
 
+
 ##
 ## Display some help about this tool.
 ##
 
-proc "help IP Discover" {list} {
-    ined browse "Help about IP Discover" {
+proc "Help IP-Discover" {list} {
+    ined browse "Help about IP-Discover" {
 	"Discover IP Network:"  
 	"    Try to discover the structure of IP networks. You should"  
 	"    use 3 Byte addresses like a.b.c or perhaps two byte addresses" 
@@ -1392,14 +1714,19 @@ proc "help IP Discover" {list} {
 	"Discover Parameter:"  
 	"    Set some parameters that control the discovering process."  
 	""  
-	"Find RPC Server:"  
+	"Show RPC Server:"  
 	"    Ask all selected hosts if a given RPC service is registered by"  
 	"    the portmapper. This can be used to locate NFS server etc."  
 	""  
-	"Find TCP Server:"  
+	"Show TCP Server:"  
 	"    Test all selected hosts if a given TCP service is available by"  
 	"    connecting to the service port. This can be used to locate"  
 	"    X11 server etc."  
+	""  
+	"Resolve Email:" 
+	"    Take an email address and resolve it to a set of hosts that" 
+	"    host a recipient. Make a traceroute to every discovered node" 
+	"    if Email Discover Routes is set to true." 
 	"" 
     }
 }
@@ -1408,9 +1735,9 @@ proc "help IP Discover" {list} {
 ## Delete the discovering tool.
 ##
 
-proc "delete IP Discover" {list} {
-    global tools
-    foreach id $tools { ined delete $id }
+proc "Delete IP-Discover" {list} {
+    global menus
+    foreach id $menus { ined delete $id }
     exit
 }
 
@@ -1418,8 +1745,9 @@ proc "delete IP Discover" {list} {
 ## Register the tool and fall into the main loop.
 ##
 
-set tools [ ined create TOOL "IP Discover" \
+set menus [ ined create MENU "IP-Discover" \
 	    "Discover IP Network" "Discover Route" "" \
-	    "Find RPC Server" "Find TCP Server" "" \
-	    "set parameter" "" \
-	    "help IP Discover" "delete IP Discover" ]
+	    "Show RPC Server" "Show TCP Server" "" \
+	    "Resolve Email" "" \
+	    "Set Parameter" "" \
+	    "Help IP-Discover" "Delete IP-Discover" ]
